/*
 * uart_lib.c
 *
 *  Created on: 11.02.2023
 *      Author: Harry L.
 *
 *  Ver. 1.1
 *
 */

#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include "main.h"
#include "uart_lib.h"

#define MAGIC_INIT_MARKER	0x31415926

typedef struct
{
	UART_HandleTypeDef *huart;
	UART_com_t *Serial_COM;
} huart_assoc_t;

static struct
{
	uint32_t init_marker; // 0x31415926
	huart_assoc_t hassoc[MAX_COM_PORTS + 1];
} huart_assoc;

#if (USE_HAL_UART_REGISTER_CALLBACKS == 1)
static void UART_RxCpltCallback(UART_HandleTypeDef *huart);
static void UART_TxCpltCallback(UART_HandleTypeDef *huart);
#endif

static bool fifo_put(UART_com_t *self, uint8_t data);
static bool fifo_get(UART_com_t *self, uint8_t *data);

/********************************************************************
 * UART API-Fuctions
 ********************************************************************/

void uart_init(UART_com_t *UART_com)
{
	int i;
	if (huart_assoc.init_marker != MAGIC_INIT_MARKER)	// only on first call
	{					// uninitialized assoc_struct. do initialization
		huart_assoc.init_marker = MAGIC_INIT_MARKER;
		for (i = 0; i < MAX_COM_PORTS + 1; i++)
		{
			huart_assoc.hassoc[i].Serial_COM = NULL;
			huart_assoc.hassoc[i].huart = NULL;
		}
	}

	for (i = 0;
			huart_assoc.hassoc[i].huart != UART_com->huart
					&& huart_assoc.hassoc[i].huart != NULL && i < MAX_COM_PORTS;
			i++)
		;
	huart_assoc.hassoc[i].huart = UART_com->huart;
	huart_assoc.hassoc[i].Serial_COM = UART_com;

	UART_com->Instance.uart_idle = true;
	UART_com->RX_Buffer[0] = 0;
	UART_com->Instance.buffer_p = 0;
	UART_com->Instance.fifo_get = 0;
	UART_com->Instance.fifo_put = 0;

#if (USE_HAL_UART_REGISTER_CALLBACKS == 1)
	UART_com->huart->RxCpltCallback = &UART_RxCpltCallback;
	UART_com->huart->TxCpltCallback = &UART_TxCpltCallback;
#endif

	HAL_UART_Receive_IT(UART_com->huart, &UART_com->Instance.rxbuf, 1);
}

void uart_put_byte(UART_com_t *UART_com, uint8_t c)
{
	static uint8_t ci;

	while (!fifo_put(UART_com, c))
		;
	if (UART_com->Instance.uart_idle == true)
	{
		UART_com->Instance.uart_idle = false;
		fifo_get(UART_com, &ci);
		HAL_UART_Transmit_IT(UART_com->huart, &ci, 1);
	}
}

void uart_put_data(UART_com_t *UART_com, uint8_t *s, size_t cnt)
{
	for (int i = 0; i < cnt; i++)
		uart_put_byte(UART_com, s[i]);
}

void uart_putc(UART_com_t *UART_com, char c)
{
	uart_put_byte(UART_com, (uint8_t) c);
}

void uart_puts(UART_com_t *UART_com, char *s)
{
	uart_put_data(UART_com, (uint8_t*) s, strlen(s));
}

/********************************************************************
 * UART Iterrupt-Handlers
 ********************************************************************/

#if (USE_HAL_UART_REGISTER_CALLBACKS == 0)
void HAL_UART_RxCpltCallback(UART_HandleTypeDef *huart)
#else
static void UART_RxCpltCallback(UART_HandleTypeDef *huart)
#endif
//static void UART_RxCpltCallback(void)
{
	UART_com_t *UART_com;
	int i;

	for (i = 0;
			huart_assoc.hassoc[i].huart != huart
					&& huart_assoc.hassoc[i].huart != NULL && i < MAX_COM_PORTS;
			i++)
		;
	UART_com = huart_assoc.hassoc[i].Serial_COM;

	if (huart == UART_com->huart)
	{
		if (UART_com->callback)
		{
			switch (UART_com->data_type)
			{
			case raw:
				UART_com->ByteCpltCallback(UART_com->Instance.rxbuf);
				break;

			case string:
				if (UART_com->Instance.rxbuf >= ' ')
				{
					if (UART_com->Instance.buffer_p
							< (UART_com->RX_Buffer_Size - 1))
					{
						UART_com->RX_Buffer[UART_com->Instance.buffer_p++] =
								UART_com->Instance.rxbuf;
						UART_com->RX_Buffer[UART_com->Instance.buffer_p] = '\0';
						if (UART_com->Echo)
							uart_putc(UART_com, UART_com->Instance.rxbuf);
					}
				}
				else
				{
					if (UART_com->Instance.rxbuf == UART_com->term_char)
					{
						// got EOL
						UART_com->RX_Buffer[UART_com->Instance.buffer_p] = 0;
						UART_com->Instance.buffer_p = 0;
						if (UART_com->Echo)
							uart_puts(UART_com, "\r\n");
						if (UART_com->StringCpltCallback)
							UART_com->StringCpltCallback(UART_com->RX_Buffer);
						UART_com->Instance.buffer_p = 0;// clear Line after Processing
					}
					else
					{
						if (UART_com->Echo)
						{
							switch (UART_com->Instance.rxbuf)
							{

							case '\b':			// backspace
								if (UART_com->Instance.buffer_p)
								{
									uart_puts(UART_com, "\b \b");
									UART_com->Instance.buffer_p--;
								}
								else
									uart_putc(UART_com, 0x07);	// beep
								break;

							case '\e':			// escape clear line
								UART_com->Instance.buffer_p = 0;
								// send singel escape
								break;
							default:
								break;
							}
						}
					}
				}
				break;

			case data:
				if (UART_com->Instance.buffer_p
						< (UART_com->RX_Buffer_Size - 1))
				{
					UART_com->RX_Buffer[UART_com->Instance.buffer_p++] =
							UART_com->Instance.rxbuf;
					if (UART_com->Instance.rxbuf == UART_com->term_char)
					{
						UART_com->DataCpltCallback(UART_com->RX_Buffer,
								UART_com->Instance.buffer_p);
						UART_com->Instance.buffer_p = 0;// clear Line after Processing
					}
				}

				break;
			}
		}
		HAL_UART_Receive_IT(UART_com->huart, &UART_com->Instance.rxbuf, 1);
	}
}

#if (USE_HAL_UART_REGISTER_CALLBACKS == 0)
void HAL_UART_TxCpltCallback(UART_HandleTypeDef *huart)
#else
void UART_TxCpltCallback(UART_HandleTypeDef *huart)
#endif
{
	static uint8_t c;
	int i;
	UART_com_t *UART_com;

	for (i = 0;
			huart_assoc.hassoc[i].huart != huart
					&& huart_assoc.hassoc[i].huart != NULL && i < MAX_COM_PORTS;
			i++)
		;
	UART_com = huart_assoc.hassoc[i].Serial_COM;

	if (huart == UART_com->huart)
	{
		if (fifo_get(UART_com, &c))
		{
			HAL_UART_Transmit_IT(UART_com->huart, (uint8_t*) &c, 1);
		}
		else
			UART_com->Instance.uart_idle = true;
	}
}

/****************************************************************
 * Implementation of a simple FIFO
 ****************************************************************/

static bool fifo_put(UART_com_t *self, uint8_t data)
{

	size_t nextput = self->Instance.fifo_put + 1;

	if (nextput == self->TX_FiFo_Size)
	{
		nextput = 0;
	}
	if (nextput == self->Instance.fifo_get)	// FIFO is full
	{
		return false;
	}
	else
	{
		self->TX_FiFo[self->Instance.fifo_put] = data;
		self->Instance.fifo_put = nextput;
		return true;
	}
}

static bool fifo_get(UART_com_t *self, uint8_t *data)
{

	if (self->Instance.fifo_get == self->Instance.fifo_put) // FIFO is empty
	{
		return false;
	}
	else
	{
		*data = self->TX_FiFo[self->Instance.fifo_get++];
		if (self->Instance.fifo_get == self->TX_FiFo_Size)
		{
			self->Instance.fifo_get = 0;
		}
		return true;
	}
}
